@@ -6,15 +6,6 @@ has _dbh => ( is => 'ro', isa => InstanceOf ['DBI::db'], init_arg => undef );
 
 has async => ( is => 'ro', isa => Bool, default => 0, init_arg => undef );
 
-# TODO requirede by autoload, WHY???
-sub DESTROY ($self) {
-    return;
-}
-
-sub dbh ($self) {
-    return $self;
-}
-
 # DBI
 sub _execute {
     my ( $self, $on_finish, $query ) = splice @_, 0, 3, ();
@@ -26,72 +17,105 @@ sub _execute {
         cb => $cb,
     );
 
-    my $dbh = $self->dbh;
+    my $exec = sub ( $dbh, $async ) {
 
-    my $async = !defined(wantarray) && $cb && $dbh->{async};
+        # prepare query
+        my $query_ref = ref $query;
 
-    # prepare query
-    my $query_ref = ref $query;
-
-    if ( !$query_ref ) {
-        $args{cache} //= 0;
-    }
-    elsif ( $query_ref eq 'ARRAY' ) {
-        $args{cache} //= 0;
-
-        $query = $self->query( $query->@* );
-
-        $bind //= $query->bind;
-    }
-    elsif ( $query_ref eq 'DBI::st' ) {
-        $args{cache} //= 0;
-    }
-    else {
-
-        # query object
-        if ( !$bind && $query->bind->@* ) {
-
-            # do not cache query by default, when query bind params are used
+        if ( !$query_ref ) {
+            $args{cache} //= 0;
+        }
+        elsif ( $query_ref eq 'ARRAY' ) {
             $args{cache} //= 0;
 
-            $bind = $query->bind;
+            $query = $self->query( $query->@* );
+
+            $bind //= $query->bind;
+        }
+        elsif ( $query_ref eq 'DBI::st' ) {
+            $args{cache} //= 0;
         }
         else {
-            $args{cache} //= 1;
+
+            # query object
+            if ( !$bind && $query->bind->@* ) {
+
+                # do not cache query by default, when query bind params are used
+                $args{cache} //= 0;
+
+                $bind = $query->bind;
+            }
+            else {
+                $args{cache} //= 1;
+            }
         }
+
+        if ($async) {
+            $dbh->execute_async( $query, $bind, \%args, $on_finish );
+
+            return;
+        }
+        else {
+            my $sth;
+
+            # prepare sth
+            if ( $args{cache} ) {
+                if ( $query_ref eq 'DBI::st' ) {
+                    $sth = $dbh->{_dbh}->prepare_cached( $query->{Statement} );
+                }
+                else {
+                    $sth = $dbh->{_dbh}->prepare_cached("$query");
+                }
+            }
+            else {
+                if ( $query_ref eq 'DBI::st' ) {
+
+                    # prepare sth again ONLY if driver support async queries, because we don't know, is sth was prepared for async or not?
+                    $sth = $dbh->{_dbh}->prepare( $query->{Statement} ) if $dbh->{async};
+                }
+                else {
+                    $sth = $dbh->{_dbh}->prepare("$query");
+                }
+            }
+
+            my $rows = $sth->execute( $bind ? $bind->@* : () ) or die $sth->errstr;
+
+            return $on_finish->( \%args, $rows, $sth );
+        }
+    };
+
+    # called from connection
+    if ( !$self->does('Pcore::DBH') ) {
+        return $exec->( $self, $self->{async} && !defined wantarray );
     }
 
-    if ($async) {
-        $dbh->execute_async( $query, $bind, \%args, $on_finish );
-
-        return;
-    }
+    # called from handle
     else {
-        my $sth;
 
-        # prepare sth
-        if ( $args{cache} ) {
-            if ( $query_ref eq 'DBI::st' ) {
-                $sth = $dbh->{_dbh}->prepare_cached( $query->{Statement} );
+        # handle is async (pgsql)
+        if ( $self->{async} ) {
+            if ( defined wantarray ) {
+                die q[Invalid usage, can't execute query synchronously via async connection];
             }
+
+            # get connection
             else {
-                $sth = $dbh->{_dbh}->prepare_cached("$query");
+                $self->dbh(
+                    sub ($dbh) {
+                        $exec->( $dbh, 1 );
+
+                        return;
+                    }
+                );
+
+                return;
             }
         }
+
+        # handle is not async (sqlite)
         else {
-            if ( $query_ref eq 'DBI::st' ) {
-
-                # prepare sth again ONLY if driver support async queries, because we don't know, is sth was prepared for async or not?
-                $sth = $dbh->{_dbh}->prepare( $query->{Statement} ) if $dbh->{async};
-            }
-            else {
-                $sth = $dbh->{_dbh}->prepare("$query");
-            }
+            return $exec->( $self, 0 );
         }
-
-        my $rows = $sth->execute( $bind ? $bind->@* : () ) or die $sth->errstr;
-
-        return $on_finish->( \%args, $rows, $sth );
     }
 }
 
@@ -366,11 +390,9 @@ sub quote_id ( $self, @ ) {
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ## | Sev. | Lines                | Policy                                                                                                         |
 ## |======+======================+================================================================================================================|
-## |    3 | 19                   | Subroutines::ProhibitExcessComplexity - Subroutine "_execute" with high complexity score (23)                  |
+## |    3 | 10                   | Subroutines::ProhibitExcessComplexity - Subroutine "_execute" with high complexity score (28)                  |
 ## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    3 | 176, 178             | References::ProhibitDoubleSigils - Double-sigil dereference                                                    |
-## |------+----------------------+----------------------------------------------------------------------------------------------------------------|
-## |    2 | 249, 252             | ControlStructures::ProhibitPostfixControls - Postfix control "while" used                                      |
+## |    2 | 273, 276             | ControlStructures::ProhibitPostfixControls - Postfix control "while" used                                      |
 ## +------+----------------------+----------------------------------------------------------------------------------------------------------------+
 ##
 ## -----SOURCE FILTER LOG END-----
